/*
 * This file is included by eval.c
 */

#define CACHE_SIZE 0x800
#define CACHE_MASK 0x7ff
#define EXPR1(c,m) ((((c)>>3)^(m))&CACHE_MASK)

struct cache_entry {		/* method hash table. */
    ID mid;			/* method's id */
    ID mid0;			/* method's original id */
    VALUE klass;		/* receiver's class */
    VALUE origin;		/* where method defined  */
    NODE *method;
    int noex;
};

static struct cache_entry cache[CACHE_SIZE];
static int ruby_running = 0;

void
rb_clear_cache()
{
  struct cache_entry *ent, *end;

  if (!ruby_running) return;
  ent = cache; end = ent + CACHE_SIZE;
  while (ent < end) {
    ent->mid = 0;
    ent++;
  }
}

static void
rb_clear_cache_for_undef(VALUE klass, ID id)
{
  struct cache_entry *ent, *end;

  if (!ruby_running) return;
  ent = cache; end = ent + CACHE_SIZE;
  while (ent < end) {
    if (ent->origin == klass && ent->mid == id) {
      ent->mid = 0;
    }
    ent++;
  }
}

static void
rb_clear_cache_by_id(ID id)
{
  struct cache_entry *ent, *end;

  if (!ruby_running) return;
  ent = cache; end = ent + CACHE_SIZE;
  while (ent < end) {
    if (ent->mid == id) {
      ent->mid = 0;
    }
    ent++;
  }
}

void
rb_clear_cache_by_class(VALUE klass)
{
  struct cache_entry *ent, *end;

  if (!ruby_running) return;
  ent = cache; end = ent + CACHE_SIZE;
  while (ent < end) {
    if (ent->klass == klass || ent->origin == klass) {
      ent->mid = 0;
    }
    ent++;
  }
}


void
rb_add_method(VALUE klass, ID mid, NODE* node, int noex)
{
  NODE *body;

  if (NIL_P(klass)){
    klass = rb_cObject;
  }
  if (ruby_safe_level >= 4 && (klass == rb_cObject || !OBJ_TAINTED(klass))) {
    rb_raise(rb_eSecurityError, "Insecure: can't define method");
  }
  if (!FL_TEST(klass, FL_SINGLETON) &&
      node && nd_type(node) != NODE_ZSUPER &&
      (mid == rb_intern("initialize" )|| mid == rb_intern("initialize_copy"))) {
    noex = NOEX_PRIVATE | noex;
  }
  else if (FL_TEST(klass, FL_SINGLETON) && node && nd_type(node) == NODE_CFUNC &&
           mid == rb_intern("allocate")) {
    rb_warn("defining %s.allocate is deprecated; use rb_define_alloc_func()",
            rb_class2name(rb_iv_get(klass, "__attached__")));
    mid = ID_ALLOCATOR;
  }
  if (OBJ_FROZEN(klass)){
    rb_error_frozen("class/module");
  }
  rb_clear_cache_by_id(mid);
  body = NEW_METHOD(node, noex);
  st_insert(RCLASS(klass)->m_tbl, mid, (st_data_t)body);
  if (node && mid != ID_ALLOCATOR && ruby_running) {
    if (FL_TEST(klass, FL_SINGLETON)) {
      rb_funcall(rb_iv_get(klass, "__attached__"), singleton_added, 1, ID2SYM(mid));
    }
    else {
      rb_funcall(klass, added, 1, ID2SYM(mid));
    }
  }
}

void
rb_define_alloc_func(VALUE klass, VALUE (*func) _((VALUE)))
{
  Check_Type(klass, T_CLASS);
  rb_add_method(CLASS_OF(klass), ID_ALLOCATOR, NEW_CFUNC(func, 0), NOEX_PRIVATE);
}

void
rb_undef_alloc_func(VALUE klass)
{
  Check_Type(klass, T_CLASS);
  rb_add_method(CLASS_OF(klass), ID_ALLOCATOR, 0, NOEX_UNDEF);
}

static NODE*
search_method(VALUE klass, ID id, VALUE *origin)
{
  NODE *body;

  if (!klass) return 0;
  while (!st_lookup(RCLASS(klass)->m_tbl, id, (st_data_t *)&body)) {
    klass = RCLASS(klass)->super;
    if (!klass) return 0;
  }

  if (origin) *origin = klass;
  return body;
}

NODE*
rb_get_method_body(VALUE *klassp, ID *idp, int *noexp)
{
  ID id = *idp;
  VALUE klass = *klassp;
  VALUE origin;
  NODE * volatile body;
  struct cache_entry *ent;

  if ((body = search_method(klass, id, &origin)) == 0 || !body->nd_body) {
    /* store empty info in cache */
    ent = cache + EXPR1(klass, id);
    ent->klass  = klass;
    ent->origin = klass;
    ent->mid = ent->mid0 = id;
    ent->noex   = 0;
    ent->method = 0;

    return 0;
  }

  if (ruby_running) {
    /* store in cache */
    ent = cache + EXPR1(klass, id);
    ent->klass  = klass;
    ent->noex   = body->nd_noex;
    if (noexp) *noexp = body->nd_noex;
    body = body->nd_body;
    if (nd_type(body) == NODE_FBODY) {
      ent->mid = id;
      *klassp = body->nd_orig;
      ent->origin = body->nd_orig;
      *idp = ent->mid0 = body->nd_mid;
      body = ent->method = body->nd_head;
    }
    else {
      *klassp = origin;
      ent->origin = origin;
      ent->mid = ent->mid0 = id;
      ent->method = body;
    }
  }
  else {
    if (noexp) *noexp = body->nd_noex;
    body = body->nd_body;
    if (nd_type(body) == NODE_FBODY) {
      *klassp = body->nd_orig;
      *idp = body->nd_mid;
      body = body->nd_head;
    }
    else {
      *klassp = origin;
    }
  }

  return body;
}

NODE*
rb_method_node(VALUE klass, ID id)
{
  int noex;
  struct cache_entry *ent;

  ent = cache + EXPR1(klass, id);
  if (ent->mid == id && ent->klass == klass && ent->method){
    return ent->method;
  }

  return rb_get_method_body(&klass, &id, &noex);
}

static void
remove_method(VALUE klass, ID mid)
{
  NODE *body;

  if (klass == rb_cObject) {
    rb_secure(4);
  }
  if (ruby_safe_level >= 4 && !OBJ_TAINTED(klass)) {
    rb_raise(rb_eSecurityError, "Insecure: can't remove method");
  }
  if (OBJ_FROZEN(klass)) rb_error_frozen("class/module");
  if (mid == __id__ || mid == __send__ || mid == init) {
    rb_warn("removing `%s' may cause serious problem", rb_id2name(mid));
  }
  if (!st_delete(RCLASS(klass)->m_tbl, &mid, (st_data_t *)&body) ||
      !body->nd_body) {
    rb_name_error(mid, "method `%s' not defined in %s",
                  rb_id2name(mid), rb_class2name(klass));
  }
  rb_clear_cache_for_undef(klass, mid);
  if (FL_TEST(klass, FL_SINGLETON)) {
    rb_funcall(rb_iv_get(klass, "__attached__"), singleton_removed, 1, ID2SYM(mid));
  }
  else {
    rb_funcall(klass, removed, 1, ID2SYM(mid));
  }
}

void
rb_remove_method(klass, name)
    VALUE klass;
    const char *name;
{
    remove_method(klass, rb_intern(name));
}

/*
 *  call-seq:
 *     remove_method(symbol)   => self
 *  
 *  Removes the method identified by _symbol_ from the current
 *  class. For an example, see <code>Module.undef_method</code>.
 */

static VALUE
rb_mod_remove_method(int argc, VALUE *argv, VALUE mod)
{
  int i;

  for (i=0; i<argc; i++) {
    remove_method(mod, rb_to_id(argv[i]));
  }
  return mod;
}

#undef rb_disable_super
#undef rb_enable_super

void
rb_disable_super(VALUE klass, const char *name)
{
  /* obsolete - no use */
}

void
rb_enable_super(VALUE klass, const char *name)
{
  rb_warning("rb_enable_super() is obsolete");
}

static void
rb_export_method(VALUE klass, ID name, ID noex)
{
  NODE *body;
  VALUE origin;

  if (klass == rb_cObject) {
    rb_secure(4);
  }
  body = search_method(klass, name, &origin);
  if (!body && TYPE(klass) == T_MODULE) {
    body = search_method(rb_cObject, name, &origin);
  }
  if (!body || !body->nd_body) {
    print_undef(klass, name);
  }
  if (body->nd_noex != noex) {
    if (klass == origin) {
      body->nd_noex = noex;
    }
    else {
      rb_add_method(klass, name, NEW_ZSUPER(), noex);
    }
  }
}

int
rb_method_boundp(VALUE klass, ID id, int ex)
{
  struct cache_entry *ent;
  int noex;

  /* is it in the method cache? */
  ent = cache + EXPR1(klass, id);
  if (ent->mid == id && ent->klass == klass) {
    if (ex && (ent->noex & NOEX_PRIVATE))
      return Qfalse;
    if (!ent->method) return Qfalse;
    return Qtrue;
  }
  if (rb_get_method_body(&klass, &id, &noex)) {
    if (ex && (noex & NOEX_PRIVATE))
      return Qfalse;
    return Qtrue;
  }
  return Qfalse;
}

void
rb_attr(VALUE klass, ID id, int read, int write, int ex)
{
  const char *name;
  char *buf;
  ID attriv;
  int noex;
  size_t len;

  if (!ex) {
    noex = NOEX_PUBLIC;
  }
  else {
    if (SCOPE_TEST(SCOPE_PRIVATE)) {
      noex = NOEX_PRIVATE;
      rb_warning((scope_vmode == SCOPE_MODFUNC) ?
                 "attribute accessor as module_function" :
                 "private attribute?");
    }
    else if (SCOPE_TEST(SCOPE_PROTECTED)) {
      noex = NOEX_PROTECTED;
    }
    else {
      noex = NOEX_PUBLIC;
    }
  }

  if (!rb_is_local_id(id) && !rb_is_const_id(id)) {
    rb_name_error(id, "invalid attribute name `%s'", rb_id2name(id));
  }
  name = rb_id2name(id);
  if (!name) {
    rb_raise(rb_eArgError, "argument needs to be symbol or string");
  }
  len = strlen(name)+2;
  buf = ALLOCA_N(char,len);
  snprintf(buf, len, "@%s", name);
  attriv = rb_intern(buf);
  if (read) {
    rb_add_method(klass, id, NEW_IVAR(attriv), noex);
  }
  if (write) {
    rb_add_method(klass, rb_id_attrset(id), NEW_ATTRSET(attriv), noex);
  }
}

static VALUE
ruby_cbase(){
  yarv_thread_t *th = GET_THREAD();
  yarv_control_frame_t *cfp = th->cfp;
  yarv_iseq_t *iseq = 0;
  VALUE cref;

  while(YARV_CONTROL_FRAME_STACK_OVERFLOW_P(th, cfp)){
    if(!CMETHOD_INFO_P(cfp->iseq)){
      iseq = cfp->iseq;
      break;
    }
    cfp = YARV_PREVIOUS_CONTROL_FRAME(cfp);
  }

  if(iseq && iseq->klass_nest_stack){
    cref = iseq->klass_nest_stack;
  }
  else{
    cref = th->klass_nest_stack;
  }
  return rb_ary_entry(cref, -1);
}

static NODE *
ruby_cref(){
  UNSUPPORTED(ruby_cref());
  return 0;
}

void
rb_undef(VALUE klass, ID id)
{
  VALUE origin;
  NODE *body;

  if (ruby_cbase() == rb_cObject && klass == rb_cObject) {
    rb_secure(4);
  }
  if (ruby_safe_level >= 4 && !OBJ_TAINTED(klass)) {
    rb_raise(rb_eSecurityError, "Insecure: can't undef `%s'", rb_id2name(id));
  }
  rb_frozen_class_p(klass);
  if (id == __id__ || id == __send__ || id == init) {
    rb_warn("undefining `%s' may cause serious problem", rb_id2name(id));
  }
  body = search_method(klass, id, &origin);
  if (!body || !body->nd_body) {
    char *s0 = " class";
    VALUE c = klass;

    if (FL_TEST(c, FL_SINGLETON)) {
      VALUE obj = rb_iv_get(klass, "__attached__");

      switch (TYPE(obj)) {
      case T_MODULE:
      case T_CLASS:
        c = obj;
        s0 = "";
      }
    }
    else if (TYPE(c) == T_MODULE) {
      s0 = " module";
    }
    rb_name_error(id, "undefined method `%s' for%s `%s'",
                  rb_id2name(id),s0,rb_class2name(c));
  }
  rb_add_method(klass, id, 0, NOEX_PUBLIC);
  if (FL_TEST(klass, FL_SINGLETON)) {
    rb_funcall(rb_iv_get(klass, "__attached__"),
               singleton_undefined, 1, ID2SYM(id));
  }
  else {
    rb_funcall(klass, undefined, 1, ID2SYM(id));
  }
}

/*
 *  call-seq:
 *     undef_method(symbol)    => self
 *  
 *  Prevents the current class from responding to calls to the named
 *  method. Contrast this with <code>remove_method</code>, which deletes
 *  the method from the particular class; Ruby will still search
 *  superclasses and mixed-in modules for a possible receiver.
 *     
 *     class Parent
 *       def hello
 *         puts "In parent"
 *       end
 *     end
 *     class Child < Parent
 *       def hello
 *         puts "In child"
 *       end
 *     end
 *     
 *     
 *     c = Child.new
 *     c.hello
 *     
 *     
 *     class Child
 *       remove_method :hello  # remove from child, still in parent
 *     end
 *     c.hello
 *     
 *     
 *     class Child
 *       undef_method :hello   # prevent any calls to 'hello'
 *     end
 *     c.hello
 *     
 *  <em>produces:</em>
 *     
 *     In child
 *     In parent
 *     prog.rb:23: undefined method `hello' for #<Child:0x401b3bb4> (NoMethodError)
 */

static VALUE
rb_mod_undef_method(int argc, VALUE *argv, VALUE mod)
{
  int i;
  for(i=0; i<argc; i++){
    rb_undef(mod, rb_to_id(argv[i]));
  }
  return mod;
}

void
rb_alias(VALUE klass, ID name, ID def)
{
  VALUE origin;
  NODE *orig, *body, *node;
  VALUE singleton = 0;

  rb_frozen_class_p(klass);
  if (name == def) return;
  if (klass == rb_cObject) {
    rb_secure(4);
  }
  orig = search_method(klass, def, &origin);
  if (!orig || !orig->nd_body) {
    if (TYPE(klass) == T_MODULE) {
      orig = search_method(rb_cObject, def, &origin);
    }
  }
  if (!orig || !orig->nd_body) {
    print_undef(klass, def);
  }
  if (FL_TEST(klass, FL_SINGLETON)) {
    singleton = rb_iv_get(klass, "__attached__");
  }
  body = orig->nd_body;
  orig->nd_cnt++;
  if (nd_type(body) == NODE_FBODY) { /* was alias */
    def = body->nd_mid;
    origin = body->nd_orig;
    body = body->nd_head;
  }

  rb_clear_cache_by_id(name);
  if (RTEST(ruby_verbose) && st_lookup(RCLASS(klass)->m_tbl, name, (st_data_t *)&node)) {
    if (node->nd_cnt == 0 && node->nd_body) {
      rb_warning("discarding old %s", rb_id2name(name));
    }
  }
  st_insert(RCLASS(klass)->m_tbl, name,
            (st_data_t)NEW_METHOD(NEW_FBODY(body, def, origin), orig->nd_noex));
  if (singleton) {
    rb_funcall(singleton, singleton_added, 1, ID2SYM(name));
  }
  else {
    rb_funcall(klass, added, 1, ID2SYM(name));
  }
}

/*
 *  call-seq:
 *     alias_method(new_name, old_name)   => self
 *  
 *  Makes <i>new_name</i> a new copy of the method <i>old_name</i>. This can
 *  be used to retain access to methods that are overridden.
 *     
 *     module Mod
 *       alias_method :orig_exit, :exit
 *       def exit(code=0)
 *         puts "Exiting with code #{code}"
 *         orig_exit(code)
 *       end
 *     end
 *     include Mod
 *     exit(99)
 *     
 *  <em>produces:</em>
 *     
 *     Exiting with code 99
 */

static VALUE
rb_mod_alias_method(VALUE mod, VALUE newname, VALUE oldname)
{
  rb_alias(mod, rb_to_id(newname), rb_to_id(oldname));
  return mod;
}


